from mylibKIv1.datastructures.nodes.SNode import SNode

class SLL:
    def __init__(self, head=None):
        self.head = head
        self.size = 0
        self.sorted = self.is_sorted
            
    def __str__(self):
        nodes = []
        current_node = self.head
        while current_node is not None:
            nodes.append(str(current_node.value))
            current_node = current_node.next
        return "->".join(nodes)
    
    def InsertHead(self, node):
        if not isinstance(node, SNode):
            raise TypeError("node argument must be an SNode object")
        node.next = self.head
        self.head = node
        self.size += 1
        
    def InsertTail(self, node):
        if not isinstance(node, SNode):
            raise TypeError("node argument must be an SNode object")
        if self.head is None:
            self.head = node
        else:
            current_node = self.head
            while current_node.next is not None:
                current_node = current_node.next
            current_node.next = node
        self.size += 1
            
    def Insert(self, node, position):
        if not isinstance(node, SNode):
            raise TypeError("node argument must be an SNode object")
        if position < 0:
            raise ValueError("position argument must be non-negative")
        if position == 0:
            self.InsertHead(node)
        else:
            current_node = self.head
            current_position = 0
            while current_node is not None and current_position < position - 1:
                current_node = current_node.next
                current_position += 1
            if current_node is None:
                raise ValueError("position argument out of range")
            node.next = current_node.next
            current_node.next = node
        self.size += 1


    def SortedInsert(self, node):
        if self.is_empty() or node.get_data() < self.head.get_data():
            self.InsertHead(node)
        else:
            current = self.head
            while current.get_next() is not None and current.get_next().get_data() < node.get_data():
                current = current.get_next()
            node.set_next(current.get_next())
            current.set_next(node)
        self.size += 1

    def Search(self, node):
        current = self.head
        while current is not None:
            if current.get_data() == node.get_data():
                return current
            current = current.get_next()
        return None
    
    def DeleteHead(self):
        if not self.is_empty():
            self.head = self.head.get_next()
        self.size -= 1  

    def DeleteTail(self):
        if not self.is_empty():
            if self.head.get_next() is None:
                self.DeleteHead()
            else:
                current = self.head
                while current.get_next().get_next() is not None:
                    current = current.get_next()
                current.set_next(None)
        self.size -= 1   
    
    def Delete(self, node):
        if not self.is_empty():
            if self.head.get_data() == node.get_data():
                self.DeleteHead()
            else:
                current = self.head
                while current.get_next() is not None and current.get_next().get_data() != node.get_data():
                    current = current.get_next()
                if current.get_next() is not None:
                    current.set_next(current.get_next().get_next())
                else:
                    print("Node not found in the list.")
        self.size -= 1
   
    def Sort(self):
        if not self.is_empty() and not self.is_sorted():
            sorted_list = SLL()
            current = self.head
            while current is not None:
                node = SNode(current.get_data())
                sorted_list.SortedInsert(node)
                current = current.get_next()
            self.head = sorted_list.head  

    def Clear(self):
        self.head = None
        self.size = 0
        self.sorted = True

    def Print(self):
        self.is_sorted()
        print(f"List length: {self.size}")
        print(f"Sorted status: {'Sorted' if self.sorted else 'Unsorted'}")
        print("List content:")
        current = self.head
        while current:
            print(current.get_data())
            current = current.get_next()   

#   HELPER FUNCTIONS
    def __str__(self):
        nodes = []
        current_node = self.head
        while current_node is not None:
            nodes.append(str(current_node.value))
            current_node = current_node.next
        return "->".join(nodes)

    
    def is_empty(self):
        return self.head is None
    
    def size(self):
        count = 0
        current = self.head
        while current is not None:
            count += 1
            current = current.get_next()
        return count
    
    def display(self):
        current = self.head
        while current is not None:
            print(current.get_data(), end=' -> ')
            current = current.get_next()
        print("None")
    
    def is_sorted(self):
        current = self.head
        while current is not None and current.get_next() is not None:
            if current.get_data() > current.get_next().get_data():
                self.sorted = False
                return False
            current = current.get_next()
        self.sorted = True
        return True



# OTHER FUNCTIONS, left in just in case
    # def Insert(self, node, position):
    #     if position == 0:
    #         self.insert_head(node)
    #     else:
    #         current = self.head
    #         count = 1
    #         while count < position and current is not None:
    #             current = current.get_next()
    #             count += 1
    #         if current is not None:
    #             node.set_next(current.get_next())
    #             current.set_next(node)
    #         else:
    #             print("Position out of range.")
    
 
    # def SortedInsert(self, node):
    #     if not isinstance(node, SNode):
    #         raise TypeError("node argument must be an SNode object")
    #     if self.head is None or node.value < self.head.value:
    #         self.InsertHead(node)
    #     else:
    #         current_node = self.head
    #         while current_node.next is not None and node.value > current_node.next.value:
    #             current_node = current_node.next
    #         node.next = current_node.next
    #         current_node.next = node
    #     self.size += 1




from mylibKIv1.datastructures.nodes.SNode import SNode
from mylibKIv1.datastructures.linear.SLL import SLL



class CSLL(SLL):
    def __init__(self, head=None):
        super().__init__(head)

    def InsertTail(self, node):
        if not self.head:
            self.head = node
            self.head.set_next(self.head)
        else:
            current = self.head
            while current.get_next() != self.head:
                current = current.get_next()
            current.set_next(node)
            node.set_next(self.head)
        self.size += 1
    
    def InsertHead(self, node):
        if not isinstance(node, SNode):
            raise TypeError("node argument must be an SNode object")
        if self.size == 0:
            node.next = node
        else:
            node.next = self.head
        self.head = node
        self.size += 1

    def Delete(self, node):
        if not self.head:
            return
        if self.head == node:
            if self.head.get_next() == self.head:
                self.head = None
            else:
                self.head = self.head.get_next()
            self.size -= 1
            return
        prev = self.head
        current = self.head.get_next()
        while current != self.head:
            if current == node:
                prev.set_next(current.get_next())
                self.size -= 1
                return
            prev = current
            current = current.get_next()
    
    def display(self):
        current = self.head
        count = 0
        while current is not None:
            print(current.get_data(), end=' -> ')
            current = current.get_next()
            count += 1
            if count >= self.size:
                break
            
        print("Repeat")

    def Print(self):
        self.is_sorted()
        print(f"List length: {self.size}")
        print(f"Sorted status: {'Sorted' if self.sorted else 'Unsorted'}")
        print("List content:")
        current = self.head
        count = 0
        while current:
            print(current.get_data())
            current = current.get_next()
            count += 1
            if count >= self.size:
                break
        
    def Sort(self):
        if not self.is_empty() and not self.is_sorted():
            sorted_list = SLL()
            current = self.head
            count = 0
            while current is not None:
                if count >= self.size:
                    break
                node = SNode(current.get_data())
                sorted_list.SortedInsert(node)
                current = current.get_next()

                count += 1
            self.head = sorted_list.head


    def SortedInsert(self, node):
        if self.is_empty() or node.get_data() < self.head.get_data():
            self.insert_head(node)
        else:
            current = self.head
            while current.get_next() is not None and current.get_next().get_data() < node.get_data():
                current = current.get_next()
            node.set_next(current.get_next())
            current.set_next(node)
        self.size += 1



class SNode:
    def __init__(self, data=None, next=None):
        self.data = data
        self.next = next

    def get_data(self):
        return self.data

    def set_data(self, data):
        self.data = data

    def get_next(self):
        return self.next

    def set_next(self, next):
        self.next = next




class TNode:
    def __init__(self, data=None, balance=0, P=None, L=None, R=None) -> None:
        self.data = data
        self.balance = balance
        self.P = P
        self.L = L
        self.R = R

    @classmethod
    def TNode(cls, data, balance=0, P=None, L=None, R=None):
        return cls(data, balance, P, L , R)

    def setData(self, data):
        self.data = data

    def setBalance(self, balance):
        self.balance = balance

    def setP(self, P):
        self.P = P

    def setL(self, L):
        self.L = L

    def setR(self, R):
        self.R = R

    def getData(self):
        return self.data
        
    def getBalance(self):
        return self.balance
    
    def getP(self):
        return self.P
    
    def getL(self):
        return self.L
    
    def getR(self):
        return self.R

    def print(self):
        if self.balance == None:
            print("Does not exist in tree")
            return
        print(f"Data: {self.data}, Balance: {self.balance}, P: {self.P}, L: {self.L}, R: {self.R}")
    
    def toString(self):
        return str(self.data)


from mylibKIv1.datastructures.nodes.DNode import DNode


class DLL:
    def __init__(self, head=None):
        self.head = head
        self.tail = head
        self.size = 0
        self.sorted = True

    def get_tail(self):
        return self.tail

    def InsertHead(self, node):
        if not self.head:
            self.head = node
            self.tail = node
        else:
            node.set_next(self.head)
            self.head.set_next(node)
            self.head = node
        self.size += 1
        self.sorted = False



    def InsertTail(self, node):
        if not self.tail:
            self.head = node
            self.tail = node
        else:
            self.tail.set_next(node)
            node.set_prev(self.tail)
            self.tail = node
        self.size += 1
        self.sorted = False

    def Insert(self, node, position):
        if position <= 0:
            self.InsertHead(node)
        elif position >= self.size:
            self.InsertTail(node)
        else:
            current = self.head
            count = 0
            while count < position:
                current = current.get_next()
                count += 1
            node.set_next(current)
            node.set_prev(current.GetPrev())
            current.GetPrev().set_next(node)
            current.set_prev(node)
            self.size += 1
            self.sorted = False

    def SortedInsert(self, node):
        if not self.head:
            self.InsertHead(node)
        elif self.sorted and node.get_data() < self.head.get_data():
            self.InsertHead(node)
        elif self.sorted and node.get_data() > self.tail.get_data():
            self.InsertTail(node)
        else:
            current = self.head
            while current.get_next() and current.get_next().get_data() < node.get_data():
                current = current.get_next()
            if current is self.head:
                node.set_next(self.head)
                self.head.set_prev(node)
                self.head = node
            else:
                node.set_next(current.get_next())
                node.set_prev(current)
                if current.get_next() != None:
                    current.get_next().set_prev(node)
                current.set_next(node)
            self.size += 1

    def Search(self, node):
        current = self.head
        while current:
            if current.get_data() == node.get_data():
                return current
            current = current.get_next()
        return None

    def DeleteHead(self):
        if not self.head:
            return
        if self.head is self.tail:
            self.head = None
            self.tail = None
        else:
            self.head = self.head.get_next()
            self.head.set_prev(None)
        self.size -= 1

    def DeleteTail(self):
        if not self.tail:
            return
        if self.head is self.tail:
            self.head = None
            self.tail = None
        else:
            self.tail = self.tail.GetPrev()
            self.tail.set_next(None)
        self.size -= 1

    def Delete(self, node):
        current = self.head
        while current:
            if current == node:
                if current is self.head:
                    self.head = current.get_next()
                    self.head.set_prev(None)
                elif current is self.tail:
                    self.tail = current.GetPrev()
                    self.tail.set_next(None)
                else:
                    current.GetPrev().set_next(current.get_next())
                    current.get_next().set_prev(current.GetPrev())
                self.size -= 1
                return True
            current = current.get_next()
        return False

    def Sort(self):
        #this dont work, it sorts fine but it doesnt set the tail to the correct value
        sorted_list = DLL()
        sorted_list.head = self.head
        sorted_list.tail = self.tail
        while True:
            if self.is_sorted():
                sorted_list.head = self.head
                sorted_list.tail = self.tail
                break
            if not self.is_empty():
                sorted_list = DLL()
                current = self.head
                while current is not None:
                    node = DNode(current.get_data())
                    sorted_list.SortedInsert(node)
                    current = current.get_next()
                self.head = sorted_list.head
                self.tail = sorted_list.tail
        sorted_list.find_tail()
        self.tail = sorted_list.get_tail()
    
    def Clear(self):
        self.head = None
        self.tail = None
        self.size = 0
        self.sorted = True

    def Print(self):
        self.is_sorted()
        print(f"List length: {self.size}")
        print(f"Sorted status: {'Sorted' if self.sorted else 'Unsorted'}")
        print("List content:")
        current = self.head
        while current:
            print(current.get_data())
            current = current.get_next()
  

# HELPER FUNCTIONS
    def display(self):
        current = self.head
        while current is not None:
            print(current.get_data(), end=' -> ')
            current = current.get_next()
        print("None")

   
    def is_empty(self):
        return self.head is None
    
    def is_sorted(self):
        # if self.sorted:
        #     return True
        current = self.head
        while current is not None and current.get_next() is not None:
            if current.get_data() > current.get_next().get_data():
                return False
            current = current.get_next()
        self.sorted = True
        return True

   
    def find_tail(self):
        current_node = self.head
        while current_node.get_next() is not None:
            current_node = current_node.get_next()
        # set the tail to the last node
        self.tail = current_node


# Binary Search Tree
from mylibKIv1.datastructures.nodes.TNode import TNode

class BST:
    def __init__(self, root=None) -> None:
        self.root = root

    @classmethod
    def BST(cls, input):
        if isinstance(input, int):
            input = TNode.TNode(input)
        return cls(input)
    
    def setRoot(self, newRoot):
        self.root = newRoot

    def getRoot(self):
        return self.root
    
    def Insert(self, input):

        if isinstance(input, int):
            new_node = TNode.TNode(input)
        else:
            new_node = input

        if self.root is None:
            self.root = new_node
            return
    
        current_node = self.root

        while True:
            if new_node.data < current_node.data:
                if current_node.L is None:
                    new_node.setP(current_node)
                    current_node.L = new_node
                    return
                else:
                    current_node = current_node.L
            elif new_node.data > current_node.data:
                if current_node.R is None:
                    new_node.setP(current_node)
                    current_node.R = new_node
                    return
                else:
                    current_node = current_node.R
            else:
                return


    def Delete(self, val):
        self.root = self._delete(self.root, val)

    def _delete(self, cur_node, val):
        if cur_node is None:
            return cur_node

        if val < cur_node.data:
            cur_node.L = self._delete(cur_node.L, val)
        elif val > cur_node.data:
            cur_node.R = self._delete(cur_node.R, val)
        else:
            # Case 1: Node with no children
            if cur_node.L is None and cur_node.R is None:
                cur_node = None
            # Case 2: Node with one child
            elif cur_node.L is None:
                cur_node = cur_node.R
            elif cur_node.R is None:
                cur_node = cur_node.L
            # Case 3: Node with two children
            else:
                temp = self._find_min(cur_node.R)
                cur_node.data = temp.data
                cur_node.R = self._delete(cur_node.R, temp.data)
        return cur_node

    def _find_min(self, cur_node):
        while cur_node.L is not None:
            cur_node = cur_node.L
        return cur_node
            
    def Search(self, val):
        return self._search(val, self.root)

    def _search(self, val, cur_node):
        if cur_node is None:
            return TNode.TNode(data=0,balance=None)
        elif cur_node.data == val:
            return cur_node
        elif val < cur_node.data:
            return self._search(val, cur_node.L)
        else:
            return self._search(val, cur_node.R)
        
    # make print out nicer?
    def printInOrder(self):
        return self._printInOrder(self.root)

    def _printInOrder(self, cur_node):
        if cur_node is not None:
            self._printInOrder(cur_node.L)
            print(cur_node.data)
            self._printInOrder(cur_node.R)

    # make look nicer?
    def printBF(self):
        if self.root is None:
            return

        cur_level = [self.root]
        while cur_level:
            next_level = []
            for node in cur_level:
                print(node.data, end=" ")
                if node.L:
                    next_level.append(node.L)
                if node.R:
                    next_level.append(node.R)
            print()
            cur_level = next_level


# AVL tree

from mylibKIv1.datastructures.nodes.TNode import TNode
from mylibKIv1.datastructures.trees.bst import BST

class AVL(BST):

    def __init__(self, root=None) -> None:
        super().__init__(root)

    @classmethod
    def AVL(cls, input):
        if isinstance(input, int):
            input = TNode.TNode(input)
        return cls(input)

    def setRoot(self, node):
        self.root = node
        AVL.update_balances(self.root)
        AVL.rebalance(node)

    def getRoot(self):
        return self.root
    
    def Insert(self, input):
        super().Insert(input)
        AVL.update_balances(self.root)
        self.root = AVL.rebalance(self.root)

    def Delete(self, val):
        self.root = self._delete(self.root, val)

    def _delete(self, cur_node, val):
        if cur_node is None:
            return cur_node

        if val < cur_node.data:
            cur_node.L = self._delete(cur_node.L, val)
        elif val > cur_node.data:
            cur_node.R = self._delete(cur_node.R, val)
        else:
            # Case 1: Node with no children
            if cur_node.L is None and cur_node.R is None:
                cur_node = None
            # Case 2: Node with one child
            elif cur_node.L is None:
                cur_node = cur_node.R
            elif cur_node.R is None:
                cur_node = cur_node.L
            # Case 3: Node with two children
            else:
                temp = self._find_min(cur_node.R)
                cur_node.data = temp.data
                cur_node.R = self._delete(cur_node.R, temp.data)
        return cur_node

    def _find_min(self, cur_node):
        while cur_node.L is not None:
            cur_node = cur_node.L
        return cur_node
    
    def Search(self, val):
        return super().Search(val)

    def printInOrder(self):
        return super().printInOrder()

    def printBF(self):
        super().printBF()

    def rebalance(root):
        # Check if the root is not None
        if root is not None:
            # Recursively rebalance the left subtree
            AVL.rebalance(root.L)
            
            # Recursively rebalance the right subtree
            AVL.rebalance(root.R)
            
            # Check if the balance factor of the root is greater than 1 or less than -1
            if root.balance < -1:
                # If the balance factor of the left child is greater than or equal to 0, perform a right rotation
                if root.L.balance <= 0:
                    root = AVL.rotate_right(root)
                # Otherwise, perform a left-right rotation
                else:
                    root.L = AVL.rotate_left(root.L)
                    root = AVL.rotate_right(root)
            elif root.balance > 1:
                # If the balance factor of the right child is less than or equal to 0, perform a left rotation
                if root.R.balance >= 0:
                    root = AVL.rotate_left(root)
                # Otherwise, perform a right-left rotation
                else:
                    root.R = AVL.rotate_right(root.R)
                    root = AVL.rotate_left(root)
            
            # Recalculate the balance factor of the root
            AVL.update_balances(root)
            
            # Check if the rebalancing caused the root node to change
            if root.P is None:
                return root
            else:
                return root.P
            
        return None


    def rotate_left(node):
        """
        Performs a left rotation on the given node and returns the new root node.
        """
        new_root = node.R
        node.R = new_root.L
        new_root.L = node
        node.balance = node.balance + 1 - min(new_root.balance, 0)
        new_root.balance = new_root.balance + 1 + max(node.balance, 0)
        if node.R is not None:
            node.R.P = node
        if node.P is not None:
            node.P.R = new_root
        new_root.P = node.P
        node.P = new_root
        return new_root


    def rotate_right(node):
        """
        Performs a right rotation on the given node and returns the new root node.
        """
        new_root = node.L
        node.L = new_root.R
        new_root.R = node
        node.balance = node.balance - 1 - max(new_root.balance, 0)
        new_root.balance = new_root.balance - 1 + min(node.balance, 0)
        if node.L is not None:
            node.L.P = node
        # if node.P is not None:
        #     node.P.L = new_root
        new_root.P = node.P
        node.P = new_root
        return new_root
    
    def update_balances(node):
        """
        Recursively updates the balance values for an AVL tree rooted at the given node.
        """
        if node is None:
            return 0
        left_height = AVL.update_balances(node.L)
        right_height = AVL.update_balances(node.R)
        node.balance = right_height - left_height
        return max(left_height, right_height) + 1


from mylibKIv1.datastructures.nodes.SNode import SNode
from mylibKIv1.datastructures.linear.SLL import SLL


class Stack(SLL):
    def __init__(self):
        super().__init__()

    def push(self, node):
        super().InsertHead(node)

    def pop(self):
        if not self.head:
            return None
        else:
            node = self.head
            self.head = self.head.get_next()
            self.size -= 1
            return node

    def peek(self):
        return self.head

    # Override non-stack methods from singlyLL with empty body
    def InsertTail(self, node):
        pass

    def Insert(self, node, position):
        pass

    def Delete(self, node):
        pass




from mylibKIv1.datastructures.linear.SLL import SLL

class Queue(SLL):
    def __init__(self):
        super().__init__()

    def enqueue(self, node):
        super().InsertHead(node)

    def dequeue(self):
        if self.size == 1:
            node = self.head
            self.head = None
            self.size -= 1
            return node
        second_last = self.head
        last = self.head.get_next()
        while True:

            second_last = last
            
            if last.get_next() == None:
                break
            last = last.get_next()
        self.DeleteTail()
        return second_last

    # Overriding methods with empty bodies
    def InsertTail(self, node):
        pass

    def SortedInsert(self, node):
        pass

    def Insert(self, node, position):
        pass

    def SortedInsert(self, node):
        pass



class DNode:
    def __init__(self, data=None, prev=None, next=None):
        self._data = data
        self._prev = prev
        self._next = next
    
    def get_data(self):
        return self._data
    
    def set_data(self, data):
        self._data = data
        
    def get_prev(self):
        return self._prev
    
    def set_prev(self, prev):
        self._prev = prev
        
    def get_next(self):
        return self._next
    
    def set_next(self, next):
        self._next = next



from setuptools import setup

setup(name ='mylibKIv1',
    version='1.1',
    description='final project for ensf 338',
    author='Isaac Hus and Kenzie Fjestad',
    author_email='isaac.hus@ucalgary.ca, kenzie.fjestad@ucalgary.ca',
    packages=['mylibKIv1','mylibKIv1.datastructures', 'mylibKIv1.datastructures.trees', 'mylibKIv1.datastructures.linear', 'mylibKIv1.datastructures.nodes']
    )


from mylibKIv1.datastructures.nodes.DNode import DNode
from mylibKIv1.datastructures.linear.DLL import DLL


class CDLL(DLL):
    def __init__(self, head=None):
        super().__init__(head)
        if self.head:
            self.head.set_prev(self.tail)
            self.tail.set_next(self.head)

    def InsertHead(self, node):
        super().InsertHead(node)
        self.head.set_prev(self.tail)
        self.tail.set_next(self.head)

    def InsertTail(self, node):
        super().InsertTail(node)
        self.head.set_prev(self.tail)
        self.tail.set_next(self.head)

    def Insert(self, node, position):
        super().Insert(node, position)
        self.head.set_prev(self.tail)
        self.tail.set_next(self.head)

    def DeleteHead(self):
        super().DeleteHead()
        self.head.set_prev(self.tail)
        self.tail.set_next(self.head)

    def DeleteTail(self):
        super().DeleteTail()
        self.head.set_prev(self.tail)
        self.tail.set_next(self.head)

    def Delete(self, node):
        if not self.head:
            return
        if node is self.head:
            self.DeleteHead()
        elif node is self.tail:
            self.DeleteTail()
        else:
            node.GetPrev().set_next(node.GetNext())
            node.GetNext().set_prev(node.GetPrev())
            self.size -= 1

    

    def display(self):
        current = self.head
        count = 0
        while current is not None:
            print(current.get_data(), end=' -> ')
            current = current.get_next()
            count += 1
            if count >= self.size:
                break
            
        print("Repeat")

    def Print(self):
        self.is_sorted()
        print(f"List length: {self.size}")
        print(f"Sorted status: {'Sorted' if self.sorted else 'Unsorted'}")
        print("List content:")
        current = self.head
        count = 0
        while current:
            print(current.get_data())
            current = current.get_next()
            count += 1
            if count >= self.size:
                break

        
    def Sort(self):
        for i in range(self.size):
            if not self.is_empty() and not self.is_sorted():
                sorted_list = DLL()
                current = self.head
                count = 0
                while current is not None:
                    if count >= self.size:
                        break
                    node = DNode(current.get_data())
                    sorted_list.SortedInsert(node)
                    current = current.get_next()

                    count += 1
                self.head = sorted_list.head



# NOT NEEDED
    # def Sort(self):
    #     if not self.head or self.size <= 1:
    #         return
    #     current = self.head.GetNext()
    #     while current is not self.head:
    #         temp = current.GetPrev()
    #         while temp is not self.head.GetPrev() and temp.GetData() > current.GetData():
    #             temp = temp.GetPrev()
    #         if temp is not current.GetPrev():
    #             current.GetPrev().GetNext().set_prev(current.GetNext())
    #             current.GetNext().set_prev(current.GetPrev())
    #             if temp is self.head:
    #                 current.set_next(self.head)
    #                 self.head.set_prev(current)
    #                 self.head = current
    #             else:
    #                 current.GetPrev().set_next(current.GetNext())
    #                 current.GetNext().set_prev(current.GetPrev())
    #                 current.set_next(temp)
    #                 temp.set_prev(current)
    #             current = current.GetPrev()
    #         current = current.GetNext()
    #     self.sorted = True